#include "FastLED.h"
#include "Wire.h"
#include "display.h"
#include "ripple.h"


// heavily based on DemoReel100 (included in FastLED-demos), made by Mark Kriegsman
// includes some routines by Andrew Tuline (https://github.com/atuline/FastLED-Demos)

#define ARRAY_SIZE(A) (sizeof(A) / sizeof((A)[0]))

// 8x8 Matrix
Display disp;


// LED strip
#define DATA_PIN    6
#define NUM_LEDS    115
#define LED_TYPE    WS2811
#define COLOR_ORDER GRB
uint8_t dispBrightness = 10;

#define FRAMES_PER_SECOND  120
CRGB leds[NUM_LEDS];

uint8_t gBrightness = 10;
uint8_t gCurrentPatternNumber;
uint8_t gHue = 0;
uint8_t gMaxPatterns = 0;
uint8_t gDoCyclePatterns = 0;
// IR command
uint8_t gGotCmd;
uint32_t gCmd;


// external effect (own source module)
Ripple theripple = Ripple(NUM_LEDS);

// List of patterns to cycle through.  Each is defined as a separate function below.
typedef void (*SimplePatternList[])();
SimplePatternList gPatterns = {
          white, red, green, yellow, blue,
          rainbow, rainbowWithGlitter, confetti, sinelon, juggle,
          bpm, ripple, threesin, noise16, dots
        };


void setup() {
  Serial.begin(9600);
  Serial.println("Start LED");

  gCurrentPatternNumber = 0;
  gMaxPatterns = ARRAY_SIZE(gPatterns);

  disp.begin();
  disp.setMaxValues(gMaxPatterns, 25);
  disp.setBrightness(dispBrightness);
  disp.setValues(gCurrentPatternNumber, gBrightness);

  Wire.begin(9); // Start I2C Bus as Slave
  Wire.onReceive(receiveEvent);

  FastLED.addLeds<LED_TYPE,DATA_PIN,COLOR_ORDER>(leds, NUM_LEDS).setCorrection(TypicalLEDStrip);
}

void loop() {
  // call the current pattern function once, updating the 'leds' array
  gPatterns[gCurrentPatternNumber]();

  // send the 'leds' array out to the actual LED strip
  FastLED.show();
  // insert a delay to keep the framerate modest
  FastLED.delay(1000/FRAMES_PER_SECOND);

  // do some periodic updates
  EVERY_N_MILLISECONDS( 20 ) { gHue++; } // slowly cycle the "base color" through the rainbow
  if (gDoCyclePatterns) {
    EVERY_N_SECONDS( 10 ) { nextPattern(); } // change patterns periodically
  }
  EVERY_N_SECONDS( 1 ) { disp.step(); }

  if (gGotCmd) {
    handleEvent();
    gGotCmd = 0;
  }
}

void receiveEvent(int howMany) {
  int avail = Wire.available();
  if (avail >= 7) {
    uint8_t protocol, b1, b2;
    uint16_t address, c1, c2;
    uint32_t lastCmd;

    gGotCmd = 1;

    // first receive the protocoll
    // (IR_NEC, IR_PANASONIC, IR_SONY8 etc.)
    // (see https://github.com/NicoHood/IRLremote)
    protocol = Wire.read();

    // now receive the adress - 2 bytes, uint16_t
    b1 = Wire.read();
    b2 = Wire.read();
    address = b1 << 8 | b2;

    // now get the 32 bit value for the command itself
    b1 = Wire.read();
    b2 = Wire.read();
    c1 = b1 << 8 | b2;
    b1 = Wire.read();
    b2 = Wire.read();
    c2 = b1 << 8 | b2;
    lastCmd = gCmd;
    gCmd = c2 << 16 | c1;
    if (gCmd == 0xFFFF) { // NEC only?
      gCmd = lastCmd;
    }

//    Serial.print("protocol: 0x");
//    Serial.println(protocol, HEX);
//    Serial.print("address: 0x");
//    Serial.println(address, HEX);
//    Serial.print("command: 0x");
//    Serial.println(gCmd, HEX);
  }
}

void handleEvent() {
    switch (gCmd) {
      case 0xFD02: // MUTE
        gDoCyclePatterns = !gDoCyclePatterns;
        disp.setAction(gDoCyclePatterns);
        break;
      case 0xEF10: // CH UP
        nextPattern();
        break;
      case 0xE916: // CH DN
        prevPattern();
        break;
      case 0xEB14: // VOL RIGHT
        if (gBrightness >= 25) {
          setBrightness(25);
        } else {
          setBrightness(gBrightness+1);
        }
        disp.activateLower(3);
        break;
      case 0xED12: // VOL LEFT
        if (gBrightness <= 1) {
          setBrightness(1);
        } else {
          setBrightness(gBrightness-1);
        }
        disp.activateLower(3);
        break;
      case 0xEC13: // OK
        gCurrentPatternNumber = 0;
        setBrightness(10);
        break;
      case 0xFE01: // POWER
        gCurrentPatternNumber = 0;
        setBrightness(1);
        break;
      case 0xF00F: // red
        gCurrentPatternNumber = 1;
        updateDisplay();
        break;
      case 0xEE11: // green
        gCurrentPatternNumber = 2;
        updateDisplay();
        break;
      case 0xEA15: // yellow
        gCurrentPatternNumber = 3;
        updateDisplay();
        break;
      case 0xE817: // blue
        gCurrentPatternNumber = 4;
        updateDisplay();
        break;
      case 0xFC03: // 1
        gCurrentPatternNumber = 5;
        updateDisplay();
        break;
      case 0xFB04: // 2
        gCurrentPatternNumber = 6;
        updateDisplay();
        break;
      case 0xFA05: // 3
        gCurrentPatternNumber = 7;
        updateDisplay();
        break;
      case 0xF906: // 4
        gCurrentPatternNumber = 8;
        updateDisplay();
        break;
      case 0xF807: // 5
        gCurrentPatternNumber = 9;
        updateDisplay();
        break;
      case 0xF708: // 6
        gCurrentPatternNumber = 10;
        updateDisplay();
        break;
      case 0xF609: // 7
        gCurrentPatternNumber = 11;
        updateDisplay();
        break;
      case 0xF50A: // 8
        gCurrentPatternNumber = 12;
        updateDisplay();
        break;
      case 0xF40B: // 9
        gCurrentPatternNumber = 13;
        updateDisplay();
        break;
      case 0xF20D: // 0
        gCurrentPatternNumber = random(gMaxPatterns);
        updateDisplay();
        break;
//      case 0xBE41: // HOLD
//        break;
//      case 0xBD42: // SELECT
//        break;
      case 0xF10E: // ENTER
          dispBrightness = (dispBrightness+1) % 15;
          disp.setBrightness(dispBrightness);
        break;
      case 0xF30C: // LAST
          dispBrightness = (dispBrightness+14) % 15;
          disp.setBrightness(dispBrightness);
        break;
      default:
        Serial.print("IR: 0x");
        Serial.println(gCmd, HEX);
        break;
    }
}

void updateDisplay() {
  disp.setValues(gCurrentPatternNumber, gBrightness);
}

void nextPattern() {
  gCurrentPatternNumber = (gCurrentPatternNumber + 1) % gMaxPatterns;
  updateDisplay();
}

void prevPattern() {
  gCurrentPatternNumber = (gCurrentPatternNumber + gMaxPatterns - 1)  % gMaxPatterns;
  updateDisplay();
}

void setBrightness(uint8_t b) {
  gBrightness = b;
  FastLED.setBrightness(gBrightness*10);
  updateDisplay();
}

void white() {
  fill_solid(leds, NUM_LEDS, CRGB::White);
}
void red() {
  fill_solid(leds, NUM_LEDS, CRGB::Red);
}
void green() {
  fill_solid(leds, NUM_LEDS, CRGB::Green);
}
void yellow() {
  fill_solid(leds, NUM_LEDS, CRGB::Yellow);
}
void blue() {
  fill_solid(leds, NUM_LEDS, CRGB::Blue);
}

void rainbow() {
  // FastLED's built-in rainbow generator
  fill_rainbow( leds, NUM_LEDS, gHue, 7);
}

void rainbowWithGlitter() {
  // built-in FastLED rainbow, plus some random sparkly glitter
  rainbow();
  addGlitter(80);
}

void addGlitter( fract8 chanceOfGlitter) {
  if(random8() < chanceOfGlitter) {
    leds[random16(NUM_LEDS)] += CRGB::White;
  }
}

void confetti() {
  // random colored speckles that blink in and fade smoothly
  fadeToBlackBy(leds, NUM_LEDS, 10);
  int pos = random16(NUM_LEDS);
  leds[pos] += CHSV(gHue + random8(64), 200, 255);
}

void sinelon() {
  // a colored dot sweeping back and forth, with fading trails
  fadeToBlackBy(leds, NUM_LEDS, 20);
  int pos = beatsin16(13,0,NUM_LEDS);
  leds[pos] += CHSV(gHue, 255, 192);
}

void bpm() {
  // colored stripes pulsing at a defined Beats-Per-Minute (BPM)
  uint8_t BeatsPerMinute = 62;
  CRGBPalette16 palette = PartyColors_p;
  uint8_t beat = beatsin8(BeatsPerMinute, 64, 255);
  for(int i = 0; i < NUM_LEDS; i++) { //9948
    leds[i] = ColorFromPalette(palette, gHue+(i*2), beat-gHue+(i*10));
  }
}

void juggle() {
  // eight colored dots, weaving in and out of sync with each other
  fadeToBlackBy(leds, NUM_LEDS, 20);
  byte dothue = 0;
  for(int i = 0; i < 8; i++) {
    leds[beatsin16(i+7,0,NUM_LEDS)] |= CHSV(dothue, 200, 255);
    dothue += 32;
  }
}

void ripple() {
  theripple.nextFrame(leds);
}

void threesin() {
  // algorithm by Andrew Tuline (https://github.com/atuline/FastLED-Demos)
  for (int k=0; k<NUM_LEDS; k++) {
    leds[k].r = qsub8(sin8(20*k + 2*gHue), 220);
    leds[k].g = qsub8(sin8(25*k + gHue),   220);
    leds[k].b = qsub8(sin8(22*k - 3*gHue), 220);
  }
}

void noise16() {
  fill_noise16(leds, NUM_LEDS, 2, gHue*100, 3000, 3, gHue*3, 50, gHue*15);
}

void dots() {
  uint8_t middle = 0;
  uint8_t side = 0;
  uint8_t other = 0;

  uint8_t x = quadwave8(gHue%256);
  other = map(x, 0, 255, NUM_LEDS/4, NUM_LEDS/4*3); // 1/4 to 3/4 of strip.
  side = map(x, 0, 255, 0, NUM_LEDS-1); // full length of strip.
  middle = map(x, 0, 255, NUM_LEDS/3, NUM_LEDS/3*2); // 1/3 to 2/3 of strip.
  leds[middle] = CRGB::Purple;
  leds[side] = CRGB::Blue;
  leds[other] = CRGB::Aqua;
  nscale8(leds,NUM_LEDS, 224); // Fade the entire array. Or for just a few LED's, use nscale8(&leds[2], 5, fadeval);
}

